@@ -16,7 +16,7 @@
  * the License, or (at your option) any later version.
  *
  * This module is maintained by Mark Rejhon at these Email addresses:
- *      mdrejhon@magi.com
+ *      marky@magmacom.com
  *      ag115@freenet.carleton.ca
  *
  * DANG_END_MODULE
@@ -24,13 +24,20 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <sys/time.h>
 #include <termios.h>
 #include <sys/ioctl.h>
+#ifdef __linux__
 #include <linux/fs.h>
+#endif
 #include <pwd.h>
+#ifdef __NetBSD__
+#include <sys/syslimits.h>		/* PATH_MAX */
+extern int errno;
+#endif
 
 #include "config.h"
 #include "emu.h"
@@ -246,6 +253,16 @@ static int ser_close(int num)
   return (i);
 }
 
+#ifdef __NetBSD__
+#define IUCLC 0
+#define OLCUC 0
+#define OCRNL ONLCR
+#define ONOCR 0
+#define ONLRET 0
+#define OFILL 0
+#define OFDEL 0
+#define XCASE 0
+#endif
 
 /* The following function is the main initialization routine that
  * initializes the UART for ONE serial port.  This includes setting up 
@@ -301,6 +318,9 @@ static void do_ser_init(int num)
     }
   }
 
+  /* FOSSIL emulation is inactive at startup. */
+  com[num].fossil_active = FALSE;
+
   /* Flag whether to emulate RTS/CTS for DOS or let Linux do the job
    * If this is nonzero, Linux will handle RTS/CTS flow control directly.
    * DANG_FIXTHIS: This needs more work before it is implemented into
@@ -369,7 +389,9 @@ static void do_ser_init(int num)
   com[num].newset.c_cflag &= ~(HUPCL | CRTSCTS);
 #endif
 
+#ifdef __linux__
   com[num].newset.c_line = 0;
+#endif
   com[num].newset.c_cc[VMIN] = 1;
   com[num].newset.c_cc[VTIME] = 0;
   if (com[num].system_rtscts) com[num].newset.c_cflag |= CRTSCTS;
