@@ -11,7 +11,7 @@
  * the License, or (at your option) any later version.
  *
  * This module is maintained by Mark Rejhon at these Email addresses:
- *      mdrejhon@magi.com
+ *      marky@magmacom.com
  *      ag115@freenet.carleton.ca
  *
  * DANG_END_MODULE
@@ -363,7 +363,7 @@ void ser_termios(int num)
     rounddiv = DIV_50;
     baud = B50;
   }
-  s_printf("divisor 0x%x -> 0x%x\n", DIVISOR, rounddiv);
+  s_printf("divisor 0x%x -> 0x%lx\n", DIVISOR, rounddiv);
 
   /* Number of 115200ths of a second for each char to transmit,
    * This assumes 10 bits per characters, although math can easily
@@ -375,10 +375,12 @@ void ser_termios(int num)
   /* Save the newbaudrate value */
   com[num].newbaud = baud;
 
+#ifdef __linux__
   /* These are required to increase chances that 57600/115200 will work. */
   /* At least, these were needed on my system, Linux 1.2 and Libc 4.6.27 */
   com[num].newset.c_cflag &= ~CBAUD;
   com[num].newset.c_cflag |= baud;
+#endif
   
   /* The following does the actual system calls to set the line parameters */
   cfsetispeed(&com[num].newset, baud);
